!++++++++++++++++++++++++++++++++++++++++++++++
!! This file is part of an Object Oriented Programming
!! in Fortran 2003 tutorial series.
!! The tutorial series was originally posted online at
!! http://dannyvanpoucke.be/fortran-tutorials-en/
!! and a pdf of the most recent
!! version can be found on ArXiv .
!!
!! (c) Dr. Dr. Danny E.P. Vanpoucke 2015
!!
!----------------------------------------------
!++++++++++++++++++++++++++++++++++
!>\brief Module containing agent-class
!! as presented in the work of Pawel Sobkowicz:
!! "Extremism without extremists: Deffuant model with emotions"
!<---------------------------------
module SobAgentClass
    implicit none
    private ! everything is hidden

type, public :: TSobAgentClass
    private
        real :: oi        !< opinion
        real :: di        !< tolerance
        real :: mu        !< global convergence speed [0-->0.5]
        real :: dmin      !< min tolerance
        real :: dmax      !< min tolerance
        integer :: L      !< exponent for tolerance update
    contains
    private
        procedure,pass(this), public :: initialize
        procedure,pass(this), public :: getOpinion
        procedure,pass(this), public :: setOpinion
        procedure,pass(this), public :: updateOpinion
        procedure,pass(this), public :: updateTolerance
        procedure,pass(this), public :: free => FREE_SOBAGENT
end type TSobAgentClass


contains
!++++++++++++++++++++++++++++++++
!>\brief Initialize the properties of the agent.
!!
!! @param[in,out] this Object to be initialized.
!! @param[in] dmin, dmax : range of the agent tolerance [\b DEFAULT 0..1 , OPTIONAL]
!! @param[in] L : integer exponent in the tolerance definition [\b DEFAULT 1 , OPTIONAL]
!! @param[in] mu: real value controlling the convergence speed [0..0.5] [\b DEFAULT 0.25, OPTIONAL]
!<-------------------------------
subroutine initialize(this, dmin,dmax, L, mu)
    class(TSobAgentClass),intent(inout) :: this
    real, intent(in),optional :: dmin,dmax,mu
    integer, intent(in), optional :: L

    this%oi=0.0
    this%di=0.0
    this%mu=0.25
    this%dmin=0.0
    this%dmax=1.0
    this%L=1
    if (present(dmin)) this%dmin=dmin
    if (present(dmax)) this%dmax=dmax
    if (present(mu))   this%mu=mu
    if (present(L))    this%L=L

end subroutine initialize
!++++++++++++++++++++++++++++++++
!>\brief Accessor for the opinion property
!!
!! @param[in] this Object to work with.
!! \return \e opinion The opinion of the agent.
!<-------------------------------
function getOpinion(this) Result(opinion)
    class(TSobAgentClass),intent(in) :: this
    real :: opinion

    opinion=this%oi
end function getOpinion
!++++++++++++++++++++++++++++++++
!>\brief external Mutator for the opinion property...also updates the tolerance.
!!
!! @param[in] this Object to work with.
!! \return \e opinion The opinion of the agent.
!<-------------------------------
subroutine setOpinion(this,opinion)
    class(TSobAgentClass),intent(inout) :: this
    real,intent(in) :: opinion

    this%oi=opinion
    call this%updateTolerance()
end subroutine setOpinion
!++++++++++++++++++++++++++++++++
!>\brief Update the opinion of this agent.
!!
!! @param[in,out] this Agent who is listening and updating its opinion.
!! @param[in] Talker The agent doing the talking and leading to the opinion update.
!<-------------------------------
subroutine updateOpinion(this, Talker)
    class(TSobAgentClass),intent(inout) :: this
    class(TSobAgentClass),intent(in) :: Talker
    !a nifty feature of Fortran 2003 OOP is the fact that inside this module, we can freely access
    ! private properties of the TSobAgentClass, independent of which agent object the private
    !property belongs to. i.e. Talker%getOpinion() can be replaced by Talker%oi, and will not give an error/warning
    if (abs(this%oi-Talker%getOpinion())<this%di) this%oi=this%oi+this%mu*(Talker%getOpinion()-this%oi)

    call this%updateTolerance()
end subroutine updateOpinion
!++++++++++++++++++++++++++++++++
!>\brief Update the tolerance of this agent.
!!
!! @param[in,out] this Agent who is updating its tolerance.
!<-------------------------------
subroutine updateTolerance(this)
    class(TSobAgentClass),intent(inout) :: this

    this%di=this%dmax-((abs(this%oi))**this%L)*(this%dmax-this%dmin)
end subroutine updateTolerance
!++++++++++++++++++++++++++++++++
!>\brief Deallocate allocated memory
!!
!! @param[in,out] this Object for which the memory is freed.
!<-------------------------------
subroutine FREE_SOBAGENT(this)
    class(TSobAgentClass),intent(inout) :: this

end subroutine FREE_SOBAGENT


end module SobAgentClass
