!++++++++++++++++++++++++++++++++++++++++++++++
!! This file is part of an Object Oriented Programming
!! in Fortran 2003 tutorial series.
!! The tutorial series was originally posted online at
!! http://dannyvanpoucke.be/fortran-tutorials-en/
!! and a pdf of the most recent
!! version can be found on ArXiv .
!!
!! (c) Dr. Dr. Danny E.P. Vanpoucke 2015
!!
!----------------------------------------------
!++++++++++++++++++++++++++++++++++
!>\brief Module containing agent-class
!! as presented in the work of Pawel Sobkowicz:
!! "Extremism without extremists: Deffuant model with emotions"
!!
!<---------------------------------
module SobPopulationClass
    implicit none
    private ! everything is hidden

    type, public :: TSobPopulationClass
        private
        integer :: nagents         !< number of agents
        real, allocatable :: oi(:) !< opinion
        real, allocatable :: di(:) !< tolerance
        real :: mu        !< global convergence speed [0-->0.5]
        real :: dmin      !< min tolerance
        real :: dmax      !< min tolerance
        integer :: L      !< exponent for tolerance update
        integer, allocatable :: talk(:) ! index list for talkers
        integer, allocatable :: listen(:)! index list for listeners
    contains
        private
        procedure,pass(this), public :: initialize
        procedure,pass(this), private :: resetTalkListenLists
        procedure,pass(this), private :: randomizeTalkListenLists
        procedure,pass(this), private :: RemoveSelfInteraction
        procedure,pass(this), private :: getOpinion_single
        procedure,pass(this), private :: getOpinion_array
        generic, public :: getOpinion => getOpinion_single, getOpinion_array
        procedure,pass(this), public :: setOpinion
        procedure,pass(this), public :: updateOpinion
        procedure,pass(this), public :: free => FREE_SOBPOPULATION
    end type TSobPopulationClass


contains
    !++++++++++++++++++++++++++++++++
    !>\brief Initialize the properties of the population.
    !!
    !! @param[in,out] this Object to be initialized.
    !<-------------------------------
    subroutine initialize(this,nagents,mu,dmin,dmax,L)
        class(TSobPopulationClass),intent(inout) :: this
        integer, intent(in) :: nagents, L
        real,intent(in) :: mu,dmin,dmax

        call RANDOM_SEED()
        call this%free()
        this%nagents=nagents
        allocate(this%oi(1:this%nagents))
        allocate(this%di(1:this%nagents))
        allocate(this%talk(1:this%nagents))
        allocate(this%listen(1:this%nagents))
        call this%resetTalkListenLists()
        this%mu=mu
        this%dmin=dmin
        this%dmax=dmax
        this%L=L
        call RANDOM_NUMBER(this%oi(:)) ! give all agents random initial opinion
        ! provide the associate tolerance
        this%di=this%dmax-((abs(this%oi))**this%L)*(this%dmax-this%dmin) ! fortran array opperations :-)

    end subroutine initialize
    !+++++++++++++++++++++++++++++++++++++++++++
    !>\brief Reset the list of listeners/talkers.
    !!
    !! @param[in,out] this Object to work with.
    !<-------------------------------
    subroutine resetTalkListenLists(this)
        class(TSobPopulationClass),intent(inout) :: this

        integer :: nr

        forall(nr=1:this%nagents) this%talk(nr)=nr
        this%listen(:)=this%talk(:)

    end subroutine resetTalkListenLists
    !+++++++++++++++++++++++++++++++++++++++++++
    !>\brief Generate the random communication lists between talkers and listeners.
    !!
    !! @param[in,out] this Object to work with.
    !<-------------------------------
    subroutine randomizeTalkListenLists(this)
        class(TSobPopulationClass),intent(inout) :: this

        call RandomizeList(this%talk)
        call RandomizeList(this%listen)
        ! However, at this point there is still a small risk of self-interaction
        call this%RemoveSelfInteraction()

    contains
        !++++++++++++++++++++++++++++++++++++++++++++
        !>\brief Subroutine to randomize a list of integers.
        !!
        !! @param[inout] List of integer values to randomize
        !<-------------------------------------------
        subroutine RandomizeList(List)
            integer, intent(inout) :: List(:)

            integer :: nr, listsize, i, j, tmp
            real, allocatable :: x(:)

            listsize=size(List)
            allocate(x(1:listsize*2))
            call random_number(x)
            x=x*listsize
            do nr=1,listsize
                i=int(x(2*nr-1))+1
                j=int(x(2*nr))+1
                tmp=List(i)
                List(i)=List(j)
                List(j)=tmp
            end do

        end subroutine RandomizeList

    end subroutine randomizeTalkListenLists
    !++++++++++++++++++++++++++++++++++++++++++++
    !>\brief Subroutine to remove self-interaction between the talk and listen lists
    !!
    !! @param[in,out] this Object to work with.
    !<-------------------------------
    subroutine RemoveSelfInteraction(this)
        class(TSobPopulationClass),intent(inout) :: this

        integer :: nr, i, tmp
        real :: x

        do nr=1,this%nagents
            if (this%talk(nr)==this%listen(nr)) then
                call random_number(x)
                i=mod(nr+int(x*(this%nagents-1)),this%nagents)+1
                tmp=this%listen(nr)
                this%listen(nr)=this%listen(i)
                this%listen(i)=tmp
            end if
        end do

    end subroutine RemoveSelfInteraction
    !+++++++++++++++++++++++++++++++++++++++++++++
    !>\brief Small function returning the opinion of the specific agent.
    !!
    !! @param[in] this Object to work with.
    !! @param[in] AgentIndex The index of the agent of interest.
    !! \return Opinion real valued opinion of a single agent. Retuns -1, when the index is out of range.
    !<-------------------------------
    function getOpinion_single(this,AgentIndex) Result(Opinion)
        class(TSobPopulationClass),intent(in) :: this
        integer, intent(in) :: AgentIndex
        real :: opinion

        opinion=-1.0
        if ((AgentIndex>0).and.(AgentIndex<=this%nagents)) opinion = this%oi(AgentIndex)
    end function getOpinion_single
    !+++++++++++++++++++++++++++++++++++++++++++++
    !>\brief Small function returning the array of opinions of all agents.
    !!
    !! @param[in] this Object to work with.
    !! \return Opinion real valued opinion of a single agent. Retuns -1, when the index is out of range.
    !<-------------------------------
    function getOpinion_array(this) Result(Opinion)
        class(TSobPopulationClass),intent(in) :: this
        real :: opinion(1:this%nagents)

        opinion(:) = this%oi(:)
    end function getOpinion_array
    !+++++++++++++++++++++++++++++++++++++++++++++
    !>\brief Small subroutine to set the opinion of the specific agent.
    !!
    !! @param[in,out] this Object to work with.
    !! @param[in] AgentIndex The index of the agent of interest.
    !! @param[in] Opinion real valued opinion of a single agent.
    !<-------------------------------
    subroutine setOpinion(this,AgentIndex,Opinion)
        class(TSobPopulationClass),intent(inout) :: this
        integer, intent(in) :: AgentIndex
        real, intent(in) :: Opinion

        if ((AgentIndex>0).and.(AgentIndex<=this%nagents)) this%oi(AgentIndex)=Opinion
    end subroutine setOpinion
    !++++++++++++++++++++++++++++++++
    !>\brief Update the opinion of this agent.
    !!
    !! @param[in,out] this Agent who is listening and updating its opinion.
    !! @param[in] Talker The agent doing the talking and leading to the opinion update.
    !<-------------------------------
    subroutine updateOpinion(this)
        class(TSobPopulationClass),intent(inout) :: this

        integer :: nr, i, j

        call this%randomizeTalkListenLists()
        do nr=1, this%nagents
            i=this%listen(nr)
            j=this%talk(nr)
            !modify the listener
            if (abs(this%oi(i)-this%oi(j))<this%di(i)) then
                this%oi(i)=this%oi(i)+this%mu*(this%oi(j)-this%oi(i))
                this%di(i)=this%dmax-((abs(this%oi(i)))**this%L)*(this%dmax-this%dmin)
            end if
            !modify the talker...note that just as for the TAgent-class setup, the updated opinion
            !of the listener is used, however, incontrast to that implementation, the current one
            !can easily be modified to not take this update into account by storing the
            !difference 'this%oi(j)-this%oi(i)' as a local variable.
            !Note that in the larger scheme of things, this has no large qualitative influence,
            !since the current setup may only speed up convergence. The paper of Sobkowicz makes no
            !mention of this, and no implementation is provided.
            if (abs(this%oi(j)-this%oi(i))<this%di(j)) then
                this%oi(j)=this%oi(j)+this%mu*(this%oi(i)-this%oi(j))
                this%di(j)=this%dmax-((abs(this%oi(j)))**this%L)*(this%dmax-this%dmin)
            end if
        end do

    end subroutine updateOpinion
    !++++++++++++++++++++++++++++++++
    !>\brief Deallocate allocated memory
    !!
    !! @param[in,out] this Object for which the memory is freed.
    !<-------------------------------
    subroutine FREE_SOBPOPULATION(this)
        class(TSobPopulationClass),intent(inout) :: this

        if (allocated(this%oi)) deallocate(this%oi)
        if (allocated(this%di)) deallocate(this%di)
        if (allocated(this%talk)) deallocate(this%talk)
        if (allocated(this%listen)) deallocate(this%listen)

    end subroutine FREE_SOBPOPULATION


end module SobPopulationClass
