!++++++++++++++++++++++++++++++++++++++++++++++
!! This file is part of an Object Oriented Programming
!! in Fortran 2003 tutorial series.
!! The tutorial series was originally posted online at
!! http://dannyvanpoucke.be/fortran-tutorials-en/
!! and a pdf of the most recent
!! version can be found on ArXiv .
!!
!! (c) Dr. Dr. Danny E.P. Vanpoucke 2015
!!
!----------------------------------------------
module Tutorial0
    use IO;
    implicit none
    private

    public :: RunTutorial0

contains
    !+++++++++++++++++++++++++++++++++++++
    !>\brief A main program routine for the agents tutorial.
    !!
    !! This version can be used as a procedural reference version
    !! of the agent program.
    !!
    !! This is the only subroutine accessable outside this module.
    !!
    !! @param[in] nragents  number of agents in the simulation. (20.000 in Sobkowicz paper)
    !! @param[in] simsteps  number of timesteps in the simulation. (500 in Sobkowicz paper)
    !! @param[in] runs      number of simulation runs, which are averaged
    !!                      to obtain the final results.(200 in Sobkowicz paper)
    !<------------------------------------
    subroutine RunTutorial0(nragents,simsteps,runs)
        integer, intent(in) :: nragents
        integer, intent(in) :: simsteps
        integer, intent(in) :: runs
        !
        integer:: L                !< parameter of the tolerance update rule
        real :: mu                 !< parameter controling the convergence speed
        real :: dmin, dmax         !< parameter of the tolerance update rule
        real, allocatable :: oi(:) !< opinion
        real, allocatable :: di(:) !< tolerance
        !support variables
        integer, allocatable :: talk(:)  !< index list for talkers
        integer, allocatable :: listen(:)!< index list for listeners
        integer :: nr, nrstep, nrrun, mesh
        real, allocatable :: opiniondynamics(:,:)
        character(len=255) :: filename

        write(*,'(A)') "----Starting Tutorial 0 subprogram.----"
        mesh=100    ! #gridpoints/bins in the opinion space?
        mu=0.25     ! Set of global parameters, following Sobkowicz paper.
        dmin=0.0001
        dmax=0.25
        L=1

        allocate(opiniondynamics(1:mesh,0:simsteps))
        opiniondynamics=0.0

        do nrrun=1,runs ! 200 repeats to average over
            call initialize(nragents,mu,dmin,dmax,L,oi,di,talk,listen)
            do nr=1,nragents
                opiniondynamics(int(oi(nr)*mesh)+1,0)=opiniondynamics(int(oi(nr)*mesh)+1,0)+1
            end do

            do nrstep=1,simsteps ! 500 simulation steps
                call updateOpinions(oi,di,mu,dmin,dmax,L,talk,listen)
                do nr=1,nragents
                    opiniondynamics(int(oi(nr)*mesh)+1,nrstep)= opiniondynamics(int(oi(nr)*mesh)+1,nrstep)+1
                end do
            end do ! end nrstep (1 timestep)
            !clean up allocated memory (needs to be done here, since allocation
            !happens during the initialization subroutine.)
            deallocate(talk)
            deallocate(listen)
            deallocate(oi)
            deallocate(di)
        end do ! end 1 run of the ensemble
        write(*,*) "Writing results"
        filename="results_procedural.dat"
        call writeOpiniondynamics(opiniondynamics,mesh,nragents,simsteps,runs,filename)
        deallocate(opiniondynamics)
        write(*,'(A)') "-------End Tutorial 0 subprogram.------"
    end subroutine RunTutorial0
    !--------------------------------------------------!
    !          Support Subroutines/functions           !
    !--------------------------------------------------!
    !+++++++++++++++++++++++++++++++++++++++++++++++
    !>\brief Initialize all variables related to the agents and their opinion.
    !!
    !! @param[in] nagents Number of agents.
    !! @param[in] mu      Parameter setting the convergence speed.
    !! @param[in] dmin, dmax, L Parameters related to the tolerance update rule.
    !! @param[out] oi     List of opinions.
    !! @param[out] di     List of tolerances.
    !! @param[out] talk, listen Lists giving the indices of how the communication
    !!             between the agents  is organized.
    !<----------------------------------------------
    subroutine initialize(nagents,mu,dmin,dmax,L,oi,di,talk,listen)
        integer, intent(in) :: nagents, L
        real,intent(in) :: mu,dmin,dmax

        real, allocatable, intent(out) :: oi(:) !< opinion
        real, allocatable, intent(out) :: di(:) !< tolerance
        integer, allocatable, intent(out) :: talk(:) ! index list for talkers
        integer, allocatable, intent(out) :: listen(:)! index list for listeners

        call RANDOM_SEED()
        allocate(oi(1:nagents))
        allocate(di(1:nagents))
        allocate(talk(1:nagents))
        allocate(listen(1:nagents))

        call resetTalkListenLists(talk,listen)
        call RANDOM_NUMBER(oi(:)) ! give all agents random initial opinion
        ! provide the associate tolerance
        di=dmax-((abs(oi))**L)*(dmax-dmin) ! fortran array opperations :-)
    end subroutine initialize
    !+++++++++++++++++++++++++++++++++++++++++++++++
    !>\brief subroutine to which resets the communication lists.
    !!
    !! @param[in,out] talk, listen Lists giving the indices of how
    !!                the communication between the agents  is organized.
    !<----------------------------------------------
    subroutine resetTalkListenLists(talk,listen)
        integer, intent(inout) :: talk(:) ! index list for talkers
        integer, intent(inout) :: listen(:)! index list for listeners

        integer :: nr
        integer :: nagents

        nagents=size(talk)
        forall(nr=1:nagents) talk(nr)=nr
        listen(:)=talk(:)
    end subroutine resetTalkListenLists
    !+++++++++++++++++++++++++++++++++++++++++++++++
    !>\brief subroutine which randomizes the communication lists.
    !!
    !! @param[in,out] talk, listen Lists giving the indices of how
    !!                the communication between the agents  is organized.
    !<----------------------------------------------
    subroutine randomizeTalkListenLists(talk,listen)
        integer, intent(inout) :: talk(:) ! index list for talkers
        integer, intent(inout) :: listen(:)! index list for listeners

        call RandomizeList(talk)
        call RandomizeList(listen)
        ! However, at this point there is still a small risk of self-interaction
        call RemoveSelfInteraction(talk,listen)
    contains
        !++++++++++++++++++++++++++++++++++++++++++++
        !>\brief (Sub)Subroutine to do thee actual randomization.
        !!
        !! @param[in,out] List of integer values to randomize
        !<-------------------------------------------
        subroutine RandomizeList(List)
            integer, intent(inout) :: List(:)

            integer :: nr, listsize, i, j, tmp
            real, allocatable :: x(:)

            listsize=size(List)
            allocate(x(1:listsize*2))
            call random_number(x)
            x=x*listsize
            do nr=1,listsize
                i=int(x(2*nr-1))+1
                j=int(x(2*nr))+1
                tmp=List(i)
                List(i)=List(j)
                List(j)=tmp
            end do
            deallocate(x)
        end subroutine RandomizeList
        !++++++++++++++++++++++++++++++++++++++++++++
        !>\brief Subroutine to remove self-interaction between the agents.
        !!
        !! @param[in,out] talk, listen Lists giving the indices of how
        !!                the communication between the agents  is organized.
        !<-------------------------------------------
        subroutine RemoveSelfInteraction(talk,listen)
            integer, intent(inout) :: talk(:) ! index list for talkers
            integer, intent(inout) :: listen(:)! index list for listeners

            integer :: nr, i, tmp, nagents
            real :: x

            nagents=size(talk)
            do nr=1,nagents
                if (talk(nr)==listen(nr)) then
                    call random_number(x)
                    i=mod(nr+int(x*(nagents-1)),nagents)+1
                    tmp=listen(nr)
                    listen(nr)=listen(i)
                    listen(i)=tmp
                end if
            end do
        end subroutine RemoveSelfInteraction

    end subroutine randomizeTalkListenLists
    !+++++++++++++++++++++++++++++++++++++++++++++++
    !>\brief Initialize all variables related to the agents and their opinion.
    !!
    !! @param[in,out] oi     List of opinions.
    !! @param[in,out] di     List of tolerances.
    !! @param[in] mu      Parameter setting the convergence speed.
    !! @param[in] dmin, dmax, L Parameters related to the tolerance update rule.
    !! @param[out] talk, listen Lists giving the indices of how the communication
    !!             between the agents  is organized.
    !<----------------------------------------------
    subroutine updateOpinions(oi,di,mu,dmin,dmax,L,talk,listen)
        real, intent(inout) :: oi(:) !< opinion
        real, intent(inout) :: di(:) !< tolerance
        real, intent(in) :: mu, dmin,dmax !< global parameters
        integer, intent(in) :: L
        integer, intent(inout) :: talk(:) !< index list for talkers
        integer, intent(inout) :: listen(:)!< index list for listeners

        integer :: nr, i, j,nagents

        nagents=size(talk)
        call randomizeTalkListenLists(talk,listen)
        do nr=1, nagents
            i=listen(nr)
            j=talk(nr)
            !modify the listener
            if (abs(oi(i)-oi(j))<di(i)) then
                oi(i)=oi(i)+mu*(oi(j)-oi(i))
                di(i)=dmax-((abs(oi(i)))**L)*(dmax-dmin)
            end if
            !modify the talker...note that just as for the TAgent-class setup, the updated opinion
            !of the listener is used, however, incontrast to that implementation, the current one
            !can easily be modified to not take this update into account by storing the
            !difference 'this%oi(j)-this%oi(i)' as a local variable.
            !Note that in the larger scheme of things, this has no large qualitative influence,
            !since the current setup may only speed up convergence. The paper of Sobkowicz makes no
            !mention of this, and no implementation is provided.
            if (abs(oi(j)-oi(i))<di(j)) then
                oi(j)=oi(j)+mu*(oi(i)-oi(j))
                di(j)=dmax-((abs(oi(j)))**L)*(dmax-dmin)
            end if
        end do
    end subroutine updateOpinions

end module Tutorial0
