!++++++++++++++++++++++++++++++++++++++++++++++
!! This file is part of an Object Oriented Programming
!! in Fortran 2003 tutorial series.
!! The tutorial series was originally posted online at
!! http://dannyvanpoucke.be/fortran-tutorials-en/
!! and a pdf of the most recent
!! version can be found on ArXiv .
!!
!! (c) Dr. Dr. Danny E.P. Vanpoucke 2015
!!
!----------------------------------------------
module Tutorial1
    use IO;
    use SobAgentClass;
    implicit none
    private

    public :: RunTutorial1

contains
    !+++++++++++++++++++++++++++++++++++++
    !>\brief A main program routine for the agents tutorial.
    !!
    !! This version is based on TAgent objects, and provides
    !! a simple first introduction into objects.
    !!
    !! This is the only subroutine accessable outside this module.
    !!
    !! @param[in] nragents  number of agents in the simulation. (20.000 in Sobkowicz paper)
    !! @param[in] simsteps  number of timesteps in the simulation. (500 in Sobkowicz paper)
    !! @param[in] runs      number of simulation runs, which are averaged
    !!                      to obtain the final results.(200 in Sobkowicz paper)
    !<------------------------------------
    subroutine RunTutorial1(nragents,simsteps,runs)
        integer, intent(in) :: nragents
        integer, intent(in) :: simsteps
        integer, intent(in) :: runs

        type(TSobAgentClass),allocatable :: agents(:) !< array of our TAgent objects
        !support variables
        integer, allocatable:: talk(:),listen(:) !< index list for talkers & listeners
        integer ::  nr, nrstep, nrupdate, nrrun, mesh
        integer:: L
        real :: mu, dmin, dmax
        real, allocatable :: opiniondynamics(:,:)
        real :: opi
        character(len=255)::filename

        write(*,'(A)') "----Starting Tutorial 1 subprogram.----"
        call RANDOM_SEED() ! initialize random number generator
        mesh=100     ! #gridpoints/bins in the opinion space?
        mu=0.25      ! Set of global parameters, following Sobkowicz paper.
        dmin=0.0001
        dmax=0.25
        L=1

        allocate(agents(1:nragents)) ! allocate memory
        allocate(talk(1:nragents))
        allocate(listen(1:nragents))
        allocate(opiniondynamics(1:mesh,0:simsteps))
        opiniondynamics=0.0

        do nrrun=1,runs ! 200 repeats to average over
            call resetTalkListenLists(talk,listen)
            !for each run reinitialize the population
            do nr=1,nragents
                call agents(nr)%initialize(dmin,dmax,L,mu)
                call RANDOM_NUMBER(opi)
                call agents(nr)%setOpinion(opi)
            end do
            !just add to opiniondynamics, and at the end divide by the number of runs to get the average
            do nr=1,nragents
                opiniondynamics(int(agents(nr)%getOpinion()*mesh)+1,0)=opiniondynamics(int(agents(nr)%getOpinion()*mesh)+1,0)+1
            end do

            do nrstep=1,simsteps ! 500 simulation steps
                call randomizeTalkListenLists(talk,listen)
                do nrupdate=1,nragents ! each agent initiates a communication
                    call agents(listen(nrupdate))%updateOpinion(agents(talk(nrupdate)))
                    !now also the talker needs to updat his opinion accoring to the Sobkowicz paper
                    call agents(talk(nrupdate))%updateOpinion(agents(listen(nrupdate)))
                end do ! end nrupdate
                do nr=1,nragents
                    opiniondynamics(int(agents(nr)%getOpinion()*mesh)+1,nrstep)= &
                        & opiniondynamics(int(agents(nr)%getOpinion()*mesh)+1,nrstep)+1
                end do
            end do ! end nrstep (1 timestep)
        end do ! end 1 run of statistical ensemble
        opiniondynamics=opiniondynamics/runs

        write(*,*) "Writing results"
        filename="results_agents.dat"
        call writeOpiniondynamics(opiniondynamics,mesh,nragents,simsteps,runs,filename)
        !filename="results3DGr.dat"
        !call writeOpiniondynamics3DGrace(opiniondynamics,mesh,nragents,simsteps,runs,filename)
        !clean up allocated memory

        write(*,*) "sizeof 1 agent=",sizeof(agents(55))
        write(*,*) "sizeof opiniondynamics=",sizeof(opiniondynamics)

        deallocate(agents)
        deallocate(opiniondynamics)
        deallocate(talk,listen)
        write(*,'(A)') "-------End Tutorial 1 subprogram.------"
    end subroutine RunTutorial1
    !--------------------------------------------------!
    !          Support Subroutines/functions           !
    !--------------------------------------------------!
    !+++++++++++++++++++++++++++++++++++++++++++++++
    !>\brief subroutine to which resets the communication lists.
    !!
    !! @param[in,out] talk, listen Lists giving the indices of how
    !!                the communication between the agents  is organized.
    !<----------------------------------------------
    subroutine resetTalkListenLists(talk,listen)
        integer, intent(inout) :: talk(:) ! index list for talkers
        integer, intent(inout) :: listen(:)! index list for listeners

        integer :: nr
        integer :: nagents

        nagents=size(talk)
        forall(nr=1:nagents) talk(nr)=nr
        listen(:)=talk(:)
    end subroutine resetTalkListenLists
    !+++++++++++++++++++++++++++++++++++++++++++++++
    !>\brief subroutine which randomizes the communication lists.
    !!
    !! @param[in,out] talk, listen Lists giving the indices of how
    !!                the communication between the agents  is organized.
    !<----------------------------------------------
    subroutine randomizeTalkListenLists(talk,listen)
        integer, intent(inout) :: talk(:) ! index list for talkers
        integer, intent(inout) :: listen(:)! index list for listeners

        call RandomizeList(talk)
        call RandomizeList(listen)
        ! However, at this point there is still a small risk of self-interaction
        call RemoveSelfInteraction(talk,listen)
    contains
        !++++++++++++++++++++++++++++++++++++++++++++
        !>\brief (Sub)Subroutine to do thee actual randomization.
        !!
        !! @param[in,out] List of integer values to randomize
        !<-------------------------------------------
        subroutine RandomizeList(List)
            integer, intent(inout) :: List(:)

            integer :: nr, listsize, i, j, tmp
            real, allocatable :: x(:)

            listsize=size(List)
            allocate(x(1:listsize*2))
            call random_number(x)
            x=x*listsize
            do nr=1,listsize
                i=int(x(2*nr-1))+1
                j=int(x(2*nr))+1
                tmp=List(i)
                List(i)=List(j)
                List(j)=tmp
            end do
            deallocate(x)
        end subroutine RandomizeList
        !++++++++++++++++++++++++++++++++++++++++++++
        !>\brief Subroutine to remove self-interaction between the agents.
        !!
        !! @param[in,out] talk, listen Lists giving the indices of how
        !!                the communication between the agents  is organized.
        !<-------------------------------------------
        subroutine RemoveSelfInteraction(talk,listen)
            integer, intent(inout) :: talk(:) ! index list for talkers
            integer, intent(inout) :: listen(:)! index list for listeners

            integer :: nr, i, tmp, nagents
            real :: x

            nagents=size(talk)
            do nr=1,nagents
                if (talk(nr)==listen(nr)) then
                    call random_number(x)
                    i=mod(nr+int(x*(nagents-1)),nagents)+1
                    tmp=listen(nr)
                    listen(nr)=listen(i)
                    listen(i)=tmp
                end if
            end do
        end subroutine RemoveSelfInteraction

    end subroutine randomizeTalkListenLists

end module Tutorial1
