!++++++++++++++++++++++++++++++++++++++++++++++
!! This file is part of an Object Oriented Programming
!! in Fortran 2003 tutorial series.
!! The tutorial series was originally posted online at
!! http://dannyvanpoucke.be/fortran-tutorials-en/
!! and a pdf of the most recent
!! version can be found on ArXiv .
!!
!! (c) Dr. Dr. Danny E.P. Vanpoucke 2015
!!
!----------------------------------------------
module Tutorial2
    use IO;
    use SobPopulationClass;
    implicit none
    private

    public :: RunTutorial2

contains
    !+++++++++++++++++++++++++++++++++++++
    !>\brief A main program routine for the agents tutorial.
    !!
    !! This version is based on Tpopulation objects, and provides
    !! a closer link with the procedural setup.
    !!
    !! This is the only subroutine accessable outside this module.
    !!
    !! @param[in] nragents  number of agents in the simulation. (20.000 in Sobkowicz paper)
    !! @param[in] simsteps  number of timesteps in the simulation. (500 in Sobkowicz paper)
    !! @param[in] runs      number of simulation runs, which are averaged
    !!                      to obtain the final results.(200 in Sobkowicz paper)
    !<------------------------------------
    subroutine RunTutorial2(nragents,simsteps,runs)
        integer, intent(in) :: nragents
        integer, intent(in) :: simsteps
        integer, intent(in) :: runs

        type(TSobPopulationClass) :: population !< A single TPopulation object.
        !support variables
        integer :: nr, nrstep, nrrun, mesh
        integer:: L
        real :: mu, dmin, dmax
        real, allocatable :: opiniondynamics(:,:)
        integer, allocatable :: opinionarray(:)
        character(len=255) :: filename

        write(*,'(A)') "----Starting Tutorial 2 subprogram.----"
        mesh=100 ! #gridpoints/bins in the opinion space?
        mu=0.25  ! Set of global parameters, following Sobkowicz paper.
        dmin=0.0001
        dmax=0.25
        L=1

        allocate(opiniondynamics(1:mesh,0:simsteps))
        allocate(opinionarray(1:nragents))
        opiniondynamics=0.0

        do nrrun=1,runs ! 200 repeats to average over
            call population%initialize(nragents,mu,dmin,dmax,L)

            opinionarray=int((population%getOpinion())*mesh)+1
            do nr=1,nragents
                opiniondynamics(opinionarray(nr),0)=opiniondynamics(opinionarray(nr),0)+1
            end do
            !One could ofcourse get direct access to the opinion of each specific agent instead
            !do nr=1,nragents
            !    opiniondynamics(int(population%getOpinion(nr)*mesh)+1,0)=opiniondynamics(int(population%getOpinion(nr)*mesh)+1,0)+1
            !end do
            do nrstep=1,simsteps ! 500 simulation steps
                call population%updateOpinion()
                !get all opinions in one go, and do 1 array multiplication with mesh,
                ! creating an easy to use lookup table for generating our opiniondynamics histogram
                opinionarray=int((population%getOpinion())*mesh)+1
                do nr=1,nragents
                    opiniondynamics(opinionarray(nr),nrstep)= &
                        & opiniondynamics(opinionarray(nr),nrstep)+1
                end do
                !One could ofcourse get direct access to the opinion of each specific agent instead
                !do nr=1,nragents
                !    opiniondynamics(int(population%getOpinion(nr)*mesh)+1,nrstep)= &
                !             & opiniondynamics(int(population%getOpinion(nr)*mesh)+1,nrstep)+1
                !end do
            end do ! end nrstep (1 timestep)
        end do ! end 1 run of statisstical ensemble
        write(*,*) "Writing results"
        filename="results_population.dat"
        call writeOpiniondynamics(opiniondynamics,mesh,nragents,simsteps,runs,filename)
        !clean up allocated memory

        write(*,*) "sizeof 1 population=",sizeof(population)
        write(*,*) "sizeof opiniondynamics=",sizeof(opiniondynamics)


        call population%free()
        deallocate(opiniondynamics)
        deallocate(opinionarray)
        write(*,'(A)') "-------End Tutorial 2 subprogram.------"
    end subroutine RunTutorial2

end module Tutorial2
